(*
 * Graph: generic graph library
 * Copyright (C) 2004
 * Sylvain Conchon, Jean-Christophe Filliatre and Julien Signoles
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU Library General Public License version 2 for more details
 * (enclosed in the file LGPL).
 *)

(* $Id: ford_fulkerson.ml,v 1.5 2004/02/19 10:35:23 signoles Exp $ *)

module type FLOW = sig
  type label
  type t
  val max_capacity : label -> t
  val min_capacity : label -> t
  val flow : label -> t
  val add : t -> t -> t
  val sub : t -> t -> t
  val zero : t
  val max : t
  val min : t
  val compare : t -> t -> int
end

module type G = sig
  type t
  module V : Sig.HASHABLE
  module E : sig
    type t
    type label
    val src : t -> V.t
    val dst : t -> V.t
    val label : t -> label
  end
  val iter_succ_e : (E.t -> unit) -> t -> V.t -> unit
  val iter_pred_e : (E.t -> unit) -> t -> V.t -> unit
end

module Make(G: G)(F: FLOW with type label = G.E.label) = 
struct

  module Mark = struct
    module H = Hashtbl.Make(G.V)
    type mark = Plus | Minus

    let marked = H.create 997
    let unvisited = Queue.create ()
		      
    let clear () = H.clear marked

    let mem = H.mem marked
		
    let set s e tag = 
      assert (not (mem s));
      H.add marked s (e, tag);
      Queue.add s unvisited
	
    let get s : G.E.t * mark =
      let e, tag = H.find marked s in
      (match e with None -> assert false | Some e -> e), tag
	  
    exception Empty = Queue.Empty
    let next () = Queue.pop unvisited
  end
      
  module Result = struct
    module E = struct
      type t = G.E.t
      module U = Util.HTProduct(G.V)(G.V)
      let equal e1 e2 =
	U.equal (G.E.src e1, G.E.dst e1) (G.E.src e2, G.E.dst e2)
      let hash e = U.hash (G.E.src e, G.E.dst e)
    end
    module H = Hashtbl.Make(E)

    let create () = H.create 997
		   
    let find = H.find
		 
    let flow r e = 
      try
	find r e
      with Not_found ->
	let f = F.flow (G.E.label e) in
	H.add r e f;
	f
	      
    let change op r e f =
      try
	H.replace r e (op (find r e) f);
      with Not_found ->
	assert false
	  
    let grow = change F.add
    let reduce = change F.sub
  end
      
  let is_full r e = 
    F.compare (F.max_capacity (G.E.label e)) (Result.flow r e) = 0
								   
  let is_empty r e =
    F.compare F.min (Result.flow r e) = 0
					  
  let set_flow r s t a =
    let rec loop t =
      if not (G.V.equal s t) then
	let e, tag = Mark.get t in
	match tag with 
	  | Mark.Plus -> Result.grow r e a; loop (G.E.src e)
	  | Mark.Minus -> Result.reduce r e a; loop (G.E.dst e)
    in
    loop t
	  
  let grow_flow r s t a =
    let rec loop u b =
      if G.V.equal s u then begin
	set_flow r s t b;
	F.add a b
      end else
	let e, tag = Mark.get u in
	let l = G.E.label e in
	match tag with
	  | Mark.Plus -> 
	      loop 
	        (G.E.src e) 
	        (min b (F.sub (F.max_capacity l) (Result.flow r e)))
	  | Mark.Minus -> 
	      loop 
	        (G.E.dst e) 
	        (min b (F.sub (Result.flow r e) (F.min_capacity l)))
    in
    loop t F.max

  let maxflow g s t =
    let r = Result.create () in
    let succ s = 
      G.iter_succ_e
	(fun e ->
	   assert (s = G.E.src e);
	   let t = G.E.dst e in
	   if not (Mark.mem t || is_full r e) then 
	     Mark.set t (Some e) Mark.Plus)
	g s
    in
    let pred s = 
      G.iter_pred_e
	(fun e ->
	   assert (s = G.E.dst e);
	   let t = G.E.src e in
	   if not (Mark.mem t || is_empty r e) then
	     Mark.set t (Some e) Mark.Minus)
	g s
    in
    let internal_loop a =
      try
	while true do let s = Mark.next () in succ s; pred s done;
	assert false
      with Mark.Empty ->
	if Mark.mem t then grow_flow r s t a else a
    in
    let rec external_loop a =
      Mark.clear ();
      Mark.set s None Mark.Plus;
      let a' = internal_loop a in
      if a = a' then a else external_loop a'
    in
    let a = external_loop F.zero in
    (fun e -> try Result.find r e with Not_found -> F.flow (G.E.label e)), a
      
end

